#include <stdbool.h>

#pragma oss assert("version.dependencies==regions")

void initialize(long const N, long const M, long const TS, double (* restrict A)[M/TS][TS][TS], double value) {
    long const NB_I = N / TS;
    long const NB_J = M / TS;

    for (long i = 0; i < NB_I; i++) {
        for (long j = 0; j < NB_J; j++) {
            #pragma oss task out(A[i][j]) label("initialize block")
            for (long ii = 0; ii < TS; ii++) {
                for (long jj = 0; jj < TS; jj++) {
                    A[i][j][ii][jj] = value;
                }
            }
        }
    }
}


void matmul(const long N, const long M, const long P, const long TS, double (*restrict A)[M/TS][TS][TS], double (*restrict B)[P/TS][TS][TS], double (*restrict C)[P/TS][TS][TS]) {
    long const NB_I = N / TS;
    long const NB_J = P / TS;
    long const NB_K = M / TS;

    for (long k = 0; k < NB_K; k++) {
        #pragma oss task weakin([N/TS]A) weakin([M/TS]B) weakreduction(+: [N/TS]C) label("K iteration")
        for (long i = 0; i < NB_I; i++) {
            for (long j = 0; j < NB_J; j++) {
                double (* restrict A_block)[TS] = A[i][k];
                double (* restrict B_block)[TS] = B[k][j];
                double (* restrict C_block)[TS] = C[i][j];

                #pragma oss task in([TS]A_block) in ([TS]B_block) reduction(+: [TS]C_block) label("matmul block")
                for (long ii = 0; ii < TS; ii++) {
                    for (long jj = 0; jj < TS; jj++) {
                        for (long kk = 0; kk < TS; kk++) {
                            C_block[ii][jj] += A_block[ii][kk] * B_block[kk][jj];
                        }
                    }
                }
            }
        }
    }
    #pragma oss taskwait

}


bool verify(long const N, long const M, long const TS, double value, double (* restrict C)[M/TS][TS][TS]) {
    long const NB_I = N / TS;
    long const NB_J = M / TS;

    bool validates = true;

    for (long i = 0; i < NB_I; i++) {
        for (long j = 0; j < NB_J; j++) {
            #pragma oss task in(C[i][j]) label("verify block") inout(validates)
            {
                for (long ii = 0; ii < TS; ii++) {
                    for (long jj = 0; jj < TS; jj++) {
                        if (C[i][j][ii][jj] != value)
                        {
                            validates = false;
                            break;
                        }
                    }
                }
            }
        }
    }
    #pragma oss taskwait

    return validates;
}
